<'
# Copyright (c) 2012-2016 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "genom3/c/client.h"
#include "private.h"
#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"


/* --- local data ---------------------------------------------------------- */

size_t <"$comp">_varmsg_maxsize = 32768;


/* <"[--- ${comp}_init --------------------------------------------------]"> */

genom_client
genom_<"$comp">_client_init(int argc, char *argv[],
                            genom_event *ex, const void **exdetail)
{
  struct pocolibs_options opts;
  char *largv[argc+1];
  genom_state_component state;
  genom_client h;
  int i, e;

  h = malloc(sizeof(*h));
  if (!h) {
    if (ex) *ex = genom_fatal;
    if (exdetail) *exdetail = NULL;
    return NULL;
  }
  h->context.raise = genom_pocolibs_client_raise;
  h->context.raised = genom_pocolibs_client_raised;
  h->context.data = &h->context_data;

  h->context_data.h = h;
  h->context_data.ex = genom_ok;
  h->context_data.exdetail = NULL;
  h->context_data.exsize = 0;

  h->instance = NULL;
  h->csid = NULL;
  h->active_rqst = NULL;
  h->nactive = h->next = 0;

  /* must shift args so that getopt() works (argv[0] must be progname) */
  largv[0] = "<"$comp">";
  for(i=0; i<argc; i++) {
    largv[i+1] = argv[i];
  }

  if (pocolibs_getopt(h, argc+1, largv, &opts)) {
    if (ex) *ex = h->context_data.ex;
    if (exdetail) *exdetail = h->context_data.exdetail;
    free(h);
    return NULL;
  }
  <"$comp">_varmsg_maxsize = opts.varmsg_maxsize;

  if (pocolibs_init(h, <"$COMP">_CLIENT_MBOX_REPLY_SIZE)) goto error;

  /* connect to server mbox */
  h->instance = strdup(opts.instance?opts.instance:"<"$comp">");
  e = csClientInit(h->instance,
                   <"$COMP">_MAX_RQST_SIZE,
                   <"$COMP">_MAX_INTERMED_REPLY_SIZE,
                   <"$COMP">_MAX_REPLY_SIZE,
                   &h->csid);
  if (e != OK) {
    e = errnoGet();
    if (H2_SYS_ERR_FLAG(e))
      genom_syserr(&(genom_syserr_detail){ .code = e }, &h->context);
    else {
      genom_mwerr_detail d;
      h2getErrMsg(e, d.what, sizeof(d.what));
      genom_mwerr(&d, &h->context);
    }
    goto error;
  }

  /* check digest */
  genom_<"$comp">_client_genom_state_init_data(&state);
  if (genom_<"$comp">_client_genom_state_port(h, &state)) goto error;

  if (strcmp(state.digest, "<"[$component digest]">")) {
    genom_incompatible_digest_detail d;
    strcpy(d.server.version, state.version);
    strcpy(d.server.date, state.date);
    strcpy(d.client.version, "<"[$component version]">");
    strcpy(d.client.date, "<"[clock format [clock sec]]">");
    genom_incompatible_digest(&d, &h->context);
    genom_<"$comp">_client_genom_state_fini_data(&state);
    goto error;
  }
  genom_<"$comp">_client_genom_state_fini_data(&state);

  return h;

error:
  if (ex) *ex = h->context_data.ex;
  if (exdetail) *exdetail = h->context_data.exdetail;
  pocolibs_fini();
  if (h->instance) free(h->instance);
  free(h);
  return NULL;
}


/* <"[--- ${comp}_fini --------------------------------------------------]"> */

void
genom_<"$comp">_client_fini(genom_client h)
{
  if (h) {
    if (h->csid) csClientEnd(h->csid);
    pocolibs_fini();
    if (h->instance) free(h->instance);
    if (h->active_rqst) free(h->active_rqst);
    free(h);
  }
}


/* <"[--- ${comp}_context -----------------------------------------------]"> */

genom_context
genom_<"$comp">_client_context(genom_client h)
{
  return &h->context;
}


/* <"[--- ${comp}_name --------------------------------------------------]"> */

const char *
genom_<"$comp">_client_instance(genom_client h)
{
  return h->instance;
}


/* <"[--- ${comp}_eventfd -----------------------------------------------]"> */

int
genom_<"$comp">_client_eventfd(genom_client h)
{
  return pocolibs_eventfd();
}


/* <"[--- ${comp}_service_info ------------------------------------------]"> */

const struct genom_service_info *
genom_<"$comp">_client_service_info(genom_client h, int rqstid)
{
  return pocolibs_service_info(h, rqstid);
}


/* <"[--- ${comp}_done --------------------------------------------------]"> */

int
genom_<"$comp">_client_done(genom_client h, int rqstid)
{
  return pocolibs_done(h, rqstid);
}


/* <"[--- ${comp}_wait --------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_wait(genom_client h, int rqstid)
{
  return pocolibs_waitterm(h, rqstid);
}


/* <"[--- ${comp}_clean -------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_clean(genom_client h, int rqstid)
{
  return pocolibs_clean(h, rqstid);
}


/* <"[--- ${comp}_doevents -----------------------------------------------]"> */

genom_event
genom_<"$comp">_client_doevents(genom_client h)
{
  return pocolibs_doevents(h);
}


/* --- component info ------------------------------------------------------ */

static const struct genom_service_info genom_service_info[] = {
<'foreach s [$component services] {'>
  {
    .name = "<"[$s name]">",
    .input = "<"[string map {\" \\\"} [json-param $s in inout]]">",
    .output = "<"[string map {\" \\\"} [json-param $s inout out]]">",
    .meta = "<"[string map {\" \\\"} [json-param-meta $s in inout out]]">",
    .init_input =
    (genom_initfn)genom_<"$comp">_client_<"[$s name]">_init_input,
    .init_output =
    (genom_initfn)genom_<"$comp">_client_<"[$s name]">_init_output,
    .fini_input =
    (genom_finifn)genom_<"$comp">_client_<"[$s name]">_fini_input,
    .fini_output =
    (genom_finifn)genom_<"$comp">_client_<"[$s name]">_fini_output,
    .json_scan =
    (genom_json_scanfn)genom_<"$comp">_client_<"[$s name]">_json_scan,
    .json_print =
    (genom_json_printfn)genom_<"$comp">_client_<"[$s name]">_json_print,
    .send =
    (genom_request_sendfn)genom_<"$comp">_client_<"[$s name]">_rqst,
    .input_size = sizeof(struct genom_<"$comp">_<"[$s name]">_input),
    .output_size = sizeof(struct genom_<"$comp">_<"[$s name]">_output)
  },
<'}'>
  { .name = NULL, .input = NULL, .output = NULL, .meta = NULL, .send = NULL }
};

static const struct genom_port_info genom_port_info[] = {
<'foreach p [$component ports out] {'>
  {
    .name = "<"[$p name]">",
    .data = "<"[string map {\" \\\"} [json-port $p]]">",
    .meta = "<"[string map {\" \\\"} [json-port-meta $p]]">",
    .init_data =
    (genom_initfn)genom_<"$comp">_client_<"[$p name]">_init_data,
    .fini_data =
    (genom_finifn)genom_<"$comp">_client_<"[$p name]">_fini_data,
    .json_print =
    (genom_json_printfn)genom_<"$comp">_client_<"[$p name]">_json_print,
<'  if {"multiple" in [$p kind]} {'>
    .simple = 0,
    .read.multiple =
    (genom_port_multiple_readfn)genom_<"$comp">_client_<"[$p name]">_port,
<'  } else {'>
    .simple = 1,
    .read.simple =
    (genom_port_readfn)genom_<"$comp">_client_<"[$p name]">_port,
<'  }'>
    .data_size = sizeof(<"[[$p datatype] declarator]">)
  },
<'}'>
  { .name = NULL, .data = NULL, .meta = NULL,
    .init_data = NULL, .fini_data = NULL, .json_print = NULL,
    .simple = 1, .read.simple = NULL, .data_size = 0
  }
};

const struct genom_client_info genom_<"$comp">_client_info = {
  .protocol = genom_client_protocol,
  .name = "<"$comp">",
  .options = genom_pocolibs_options,

  .init = genom_<"$comp">_client_init,
  .fini = genom_<"$comp">_client_fini,
  .context = genom_<"$comp">_client_context,
  .instance = genom_<"$comp">_client_instance,
  .eventfd = genom_<"$comp">_client_eventfd,
  .service_info = genom_<"$comp">_client_service_info,
  .done = genom_<"$comp">_client_done,
  .wait = genom_<"$comp">_client_wait,
  .result = pocolibs_result,
  .clean = genom_<"$comp">_client_clean,
  .abort = genom_<"$comp">_client_abort,
  .doevents = genom_<"$comp">_client_doevents,
  .json_error = genom_<"$comp">_client_json_error,

  .nservices = <"[llength [$component services]]">,
  .services = genom_service_info,

  .nports = <"[llength [$component ports out]]">,
  .ports = genom_port_info
};

<'set x 0; foreach s [$component services] {'>
const struct genom_service_info *genom_<"$comp">_client_<"[$s name]">_info =
  &genom_service_info[<"$x">];
<'  incr x }'>

<'set x 0; foreach p [$component ports out] {'>
const struct genom_port_info *genom_<"$comp">_client_<"[$p name]">_info =
  &genom_port_info[<"$x">];
<'  incr x }'>
