<'
# Copyright (c) 2012-2015 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Aug  1 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include "commonStructLib.h"

#include "<"$comp">_control_task.h"


<'foreach obj [list $task {*}[$task services]] {'>

/* <"[--- [$obj class] [$obj name] --------------------------------------]"> */
<'  foreach state [$obj fsm] {'>
<'    set codel [$obj fsm $state]'>
<'    set func [$obj class]_[$obj name]_[$state name]'>

/* state <"[$state name]"> */
static __inline__ genom_event
<"$comp">_codel_<"$func">(
  struct genom_component_data *self, struct genom_activity *activity)
{
<'    if {[llength [$codel parameters local]]} {'>
  struct genom_<"$comp">_<"[$obj name]">_activity *a =
    (struct genom_<"$comp">_<"[$obj name]">_activity *)activity;
<'    }'>
  genom_event s;

  genom_take_resource(
    self,
<'    foreach m [$codel mutex] {'>
<'      switch [$m class] {'>
<'        codel { set rname [$m cname] }'>
<'        service { set rname genom_${comp}_[$m name]_controlcb }'>
<'        default { error "unsupported object" }'>
<'      }'>
<'      if {[catch {$m task} t]} {'>
    self->resources.control == <"$rname"> ||
<'      } else {'>
    self->resources.task_<"[$t name]"> == <"$rname"> ||
<'      }'>
<'    }'>
    self->resources.all,

    self->resources.task_<"[$task name]"> = <"[$codel cname]">);
<'
    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- [$p src] {
        local {
          if {[$obj class] eq "task"} {
            error "task codel with service parameter"
          }
          switch -- [[$p param] dir] {
            local	{ set m "a->locals.[$p name]" }
            in		{ set m "a->in.[$p name]" }
            inout - out { set m "a->out.[$p name]" }
            default	{ error "unsupported local direction" }
          }
        }
        remote	{ set m "self->remotes.[[$p remote] name].data" }
        port	{ set m "self->ports.[[$p port] name].handle" }
        ids	{ set m "self->ids[$p member]" }
        default	{ error "unsupported codel parameter source" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
        in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
        default {
          lappend plist [[$p type] pass reference $m]
        }
      }
    }
    lappend plist "&self->tasks.[$task name].context"
'>
  s = <"[$codel invoke $plist]">;
  genom_give_resource(self, self->resources.task_<"[$task name]"> = NULL);

  return s;
}
<'  }'>

/* invoke codels according to <"[$obj name]"> state */
static __inline__ enum genom_activity_status
<"$comp">_invoke_<"[$obj class]_[$obj name]">(
  struct genom_component_data *self, struct genom_activity *a)
{
  genom_event s;

  genom_log_debug(
    "<"task [$task name] invoking [$obj class] [$obj name] %s">", a->state);

<'  foreach state [$obj fsm] {'>
<'    set func [$obj class]_[$obj name]_[$state name]'>
  if (a->state == <"[$state cname]">) {
    s = <"$comp">_codel_<"$func">(self, a);
    genom_log_debug("<"[$obj class] [$obj name]"> yielded %s", s);
    if (
<'    foreach y [[$obj fsm $state] yield] {'>
<'      if {"pause" in [$y kind]} continue'>
<'      if {"ether" eq [$y name]} continue'>
      s == <"[$y cname]"> ||
<'  }'>
      0) {
      a->state = s;
      return ACT_RUN;
    }
    if (
<'    foreach y [[$obj fsm $state] yield] {'>
<'      if {"pause" in [$y kind]} { '>
      (s == <"[$y cname]"> && (s = <"$comp">_<"[$y name]">)) ||
<'      }'>
<'    }'>
      0) {
      a->state = s;
      a->pause = 1;
      return ACT_RUN;
    }
    if (
<'    foreach y [[$obj fsm $state] yield] {'>
<'      if {"ether" eq [$y name]} {'>
      s == <"[$y cname]"> ||
<'      }'>
<'    }'>
      0) {
      a->state = <"$comp">_ether;
      return ACT_ETHER;
    }
    goto ex;
  }

<'  }'>
<'  if {[$obj class] eq "task"} {'>
  /* default start, if not already handled above */
  if (a->state == <"$comp">_start) {
    a->state = <"$comp">_ether;
    return ACT_ETHER;
  }
<'  }'>

  /* default stop, if not already handled above */
  if (a->state == <"$comp">_stop) {
    genom_interrupted_detail d;

    assert(a->interruptedby);
    strncpy(d.by, a->interruptedby, sizeof(d.by));
    d.by[sizeof(d.by)-1] = *"";
    s = genom_interrupted(&d, &self->tasks.<"[$task name]">.context);
    goto ex;
  }

  /* this cannot happen by construction */
  genom_log_warn(0, "bad state (%s) in <"[$obj class]_[$obj name]">",
                 a->state?a->state : "genom_ok");
  abort();
  /*NOTREACHED*/

ex:
  if (
<'  foreach t [$obj throw] {'>
    s != <"[$t cname]">_id &&
<'  }'>
    1) {
    genom_bad_transition_detail d;
    strncpy(d.from, a->state, sizeof(d.from)); d.from[sizeof(d.from)-1] = *"";
    strncpy(d.to, s, sizeof(d.to)); d.to[sizeof(d.to)-1] = *"";
    genom_log_warn(
      0, "bad transition from %s to %s in <"[$obj class] [$obj name]">",
      a->state, s?s:"genom_ok");
    genom_bad_transition(&d, &self->tasks.<"[$task name]">.context);
  }

  a->state = NULL;
  return ACT_ETHER;
}

<'}'>


/* <"[--- task [$task name] ==============================================]"> */

enum genom_activity_status
<"genom_${comp}_[$task name]_invoke">(
  struct genom_component_data *self, struct genom_activity *a)
{
  switch(a->sid) {
    case -1:
      return <"${comp}_invoke_task_[$task name]">(self, a);

<'foreach s [$task services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return <"${comp}_invoke_[$s class]_[$s name]">(self, a);
<'}'>
  }

  /* must not happen */
  genom_log_warn(0, "bad activity %d in task <"[$task name]">", a->sid);
  abort();
  return ACT_VOID;
}
